#include <stdio.h>
#include <math.h>
#include <limits.h>
#include "oscillators.h"
#include "sndconf.h"

/* Save wave position between functions */
float phi;

/* Hard-clip indicator */
int clip = 0;

/* Hard limit volume to sample format limits */
float limit(float y)
{
  if (y < FORMAT_MIN) {
    y = FORMAT_MIN;
    if (clip == 0) {
      fprintf(stderr, "Sample format upper limit exceeded: Hard-clipping\n");
      clip = 1;
    }
  } else if (y > FORMAT_MAX) {
    y = FORMAT_MAX;
    if (clip == 0) {
      fprintf(stderr, "Sample format lower limit exceeded: Hard-clipping\n");
      clip = 1;
    }
  }
  return y;
}

/* Sine wave generator */
float sine(float A, float omega, float t)
{
  return A * sin(omega * t + phi);
}

/* Square wave generator */
float square(float A, float omega, float t)
{
  if (sin(omega * t + phi) > 0.0f)
    return A;
  else
    return -A;
} 

/* Triangle wave generator */
float triangle(float A, float omega, float t)
{
  return A * asin(sin(omega * t + phi)) / (M_PI / 2.0f);
}

/* Sawtooth wave generator */
float saw(float A, float omega, float t)
{
  /* Calculate period */
  float T = 2.0f * M_PI / omega;

/*  return (t/T - floor(t/T)) * A; */
  return ((t/T - floor(t/T)) - 0.5f) * A;
}


/* Input arg: frequency */
void osc_mono(float (*osc)(float A, float omega, float t),
	      float freq, float time, float volume)
{
  FILE *fp = NULL;

  /* Midpoint of bandwith */
  const FORMAT_TYPE mid = (FORMAT_MAX + FORMAT_MIN)/2;

  /* Amplitude */
  const float A = (float)mid * volume;

  /* Current time */
  float t;

  /* Position of wave at t */
  float y = 0;

  /* Angular frequency [rad/s] of 1st harmonic */
  const float omega = 2.0f * M_PI * freq;

  if (sndparams.write_wave == 1) {
    fp = fopen("wave.dat", "w");
  }

  for (t = 0.0f; t < time; t += 1.0/sndparams.rate) {
    y = mid;

    /* Fundamental tone */
    y += osc(A, omega, t);

    /* Write wave position to stdout for aplay */
    y = limit(y); /* Hard-clip signal to limits */
    putchar((FORMAT_TYPE)y);

    if (sndparams.write_wave == 1)
      fprintf(fp, "%f\t%f\n", t, y);
  }

  if (sndparams.write_wave == 1)
    fclose(fp);

  /* Unlock clipping indicator */
  clip = 0;

  /* Save wave position for next oscillator call */
  phi = y;
}

/* Input arg: frequency */
void osc_mono_16h(float (*osc)(float A, float omega, float t),
    		  float freq, float time, float volume)
{
  FILE *fp = NULL;

  /* Midpoint of bandwith */
  const FORMAT_TYPE mid = (FORMAT_MAX + FORMAT_MIN)/2;

  /* Amplitude */
  const float A = (float)mid * volume;

  /* Current time */
  float t;

  /* Position of wave at t */
  float y = 0;

  /* Angular frequency [rad/s] of 1st harmonic */
  const float omega = 2.0f * M_PI * freq;

  if (sndparams.write_wave == 1) {
    fp = fopen("wave.dat", "w");
  }

  /* Overtone counter */
  int o;

  for (t = 0.0f; t < time; t += 1.0/sndparams.rate) {
    y = mid;

    /* Fundamental tone */
    y += osc(A, omega, t);

    /* Create 16 orders of overtones */
    float vol = 0.2f; /* Vol. of first harmonic */
    for (o = 2; o<17; ++o) {
      y += osc(A*vol, omega*(float)o, t);
    }

    /* Write wave position to stdout for aplay */
    y = limit(y); /* Hard-clip signal to limits */
    putchar((FORMAT_TYPE)y);

    if (sndparams.write_wave == 1)
      fprintf(fp, "%f\t%f\n", t, y);
  }

  if (sndparams.write_wave == 1)
    fclose(fp);

  /* Unlock clipping indicator */
  clip = 0;

  /* Save wave position for next oscillator call */
  phi = y;
}

